/**
 * SQL-DK
 * Copyright © 2013 František Kučera (frantovo.cz)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package info.globalcode.sql.dk;

import info.globalcode.sql.dk.configuration.Configuration;
import info.globalcode.sql.dk.configuration.ConfigurationException;
import info.globalcode.sql.dk.configuration.ConfigurationProvider;
import info.globalcode.sql.dk.configuration.DatabaseDefinition;
import info.globalcode.sql.dk.configuration.FormatterDefinition;
import info.globalcode.sql.dk.formatting.ColumnsHeader;
import info.globalcode.sql.dk.formatting.Formatter;
import info.globalcode.sql.dk.formatting.FormatterContext;
import info.globalcode.sql.dk.formatting.FormatterException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.sql.rowset.RowSetMetaDataImpl;

/**
 * Displays info like help, version etc.
 *
 * @author Ing. František Kučera (frantovo.cz)
 */
public class InfoLister {

	private static final Logger log = Logger.getLogger(InfoLister.class.getName());
	private PrintStream out;
	private ConfigurationProvider configurationProvider;
	private CLIOptions options;
	private Formatter formatter;

	public InfoLister(PrintStream out, ConfigurationProvider configurationProvider, CLIOptions options) {
		this.out = out;
		this.configurationProvider = configurationProvider;
		this.options = options;
	}

	public void showInfo() throws ConfigurationException, FormatterException {
		EnumSet<InfoType> commands = options.getShowInfo();

		boolean formattinNeeded = false;

		for (InfoType infoType : commands) {
			switch (infoType) {
				case CONNECTION:
				case DATABASES:
				case FORMATTERS:
				case TYPES:
					formattinNeeded = true;
					break;
			}
		}

		if (formattinNeeded) {
			try (Formatter f = getFormatter()) {
				formatter = f;
				formatter.writeStartBatch();
				formatter.writeStartDatabase(new DatabaseDefinition());
				formatter.writeStartStatement();
				showInfos(commands);
				formatter.writeEndStatement();
				formatter.writeEndDatabase();
				formatter.writeEndBatch();
				formatter.close();
			}
		} else {
			showInfos(commands);
		}
	}

	private void showInfos(EnumSet<InfoType> commands) throws ConfigurationException, FormatterException {
		for (InfoType infoType : commands) {
			infoType.showInfo(this);
		}
	}

	private void listFormatters() throws ConfigurationException, FormatterException {
		ColumnsHeader header = constructHeader(
				new HeaderField("name", SQLType.VARCHAR),
				new HeaderField("built_in", SQLType.BOOLEAN),
				new HeaderField("default", SQLType.BOOLEAN),
				new HeaderField("class_name", SQLType.VARCHAR));
		List<Object[]> data = new ArrayList<>();

		String defaultFormatter = configurationProvider.getConfiguration().getDefaultFormatter();
		defaultFormatter = defaultFormatter == null ? Configuration.DEFAULT_FORMATTER : defaultFormatter;

		for (FormatterDefinition fd : configurationProvider.getConfiguration().getBuildInFormatters()) {
			data.add(new Object[]{fd.getName(), true, defaultFormatter.equals(fd.getName()), fd.getClassName()});
		}

		for (FormatterDefinition fd : configurationProvider.getConfiguration().getFormatters()) {
			data.add(new Object[]{fd.getName(), false, defaultFormatter.equals(fd.getName()), fd.getClassName()});
		}

		printTable(formatter, header, data);


	}

	public void listTypes() throws FormatterException, ConfigurationException {
		ColumnsHeader header = constructHeader(new HeaderField("name", SQLType.VARCHAR), new HeaderField("code", SQLType.INTEGER));
		List<Object[]> data = new ArrayList<>();
		for (SQLType sqlType : SQLType.values()) {
			data.add(new Object[]{sqlType.name(), sqlType.getCode()});
		}
		printTable(formatter, header, data);
		log.log(Level.INFO, "Type names in --types option are case insensitive");
	}

	public void listDatabases() throws ConfigurationException, FormatterException {
		ColumnsHeader header = constructHeader(
				new HeaderField("database_name", SQLType.VARCHAR),
				new HeaderField("user_name", SQLType.VARCHAR),
				new HeaderField("database_url", SQLType.VARCHAR));
		List<Object[]> data = new ArrayList<>();

		final List<DatabaseDefinition> configuredDatabases = configurationProvider.getConfiguration().getDatabases();
		if (configuredDatabases.isEmpty()) {
			log.log(Level.WARNING, "No databases are configured.");
		} else {
			for (DatabaseDefinition dd : configuredDatabases) {
				data.add(new Object[]{dd.getName(), dd.getUserName(), dd.getUrl()});
			}
		}

		printTable(formatter, header, data);
	}

	public void testConnection() throws FormatterException, ConfigurationException {
		ColumnsHeader header = constructHeader(
				new HeaderField("database_name", SQLType.VARCHAR),
				new HeaderField("configured", SQLType.BOOLEAN),
				new HeaderField("connected", SQLType.BOOLEAN));
		List<Object[]> data = new ArrayList<>();

		for (String dbName : options.getDatabaseNamesToTest()) {
			data.add(testConnection(dbName));
		}

		printTable(formatter, header, data);
	}

	public Object[] testConnection(String dbName) {
		log.log(Level.FINE, "Testing connection to database: {0}", dbName);

		boolean succesfullyConnected = false;
		boolean succesfullyConfigured = false;

		try {
			DatabaseDefinition dd = configurationProvider.getConfiguration().getDatabase(dbName);
			log.log(Level.FINE, "Database definition was loaded from configuration");
			succesfullyConfigured = true;
			try (DatabaseConnection dc = dd.connect(options.getDatabaseProperties())) {
				succesfullyConnected = dc.test();
			}
			log.log(Level.FINE, "Database connection test was successful");
		} catch (ConfigurationException | SQLException e) {
			log.log(Level.SEVERE, "Error during testing connection", e);
		}

		return new Object[]{dbName, succesfullyConfigured, succesfullyConnected};
	}

	public void printResource(String fileName) {
		try (BufferedReader reader = new BufferedReader(new InputStreamReader(getClass().getClassLoader().getResourceAsStream(fileName)))) {
			while (true) {
				String line = reader.readLine();
				if (line == null) {
					break;
				} else {
					println(line);
				}
			}
		} catch (Exception e) {
			log.log(Level.SEVERE, "Unable to print this info. Please see our website for it: " + Constants.WEBSITE, e);
		}
	}

	private void println(String line) {
		out.println(line);
	}

	private void printTable(Formatter formatter, ColumnsHeader header, List<Object[]> data) throws ConfigurationException, FormatterException {
		formatter.writeStartResultSet(header);

		for (Object[] row : data) {
			formatter.writeStartRow();
			for (Object cell : row) {
				formatter.writeColumnValue(cell);
			}
			formatter.writeEndRow();
		}

		formatter.writeEndResultSet();
	}

	private Formatter getFormatter() throws ConfigurationException, FormatterException {
		String formatterName = options.getFormatterName();
		formatterName = formatterName == null ? Configuration.DEFAULT_FORMATTER_PREFETCHING : formatterName;
		FormatterDefinition fd = configurationProvider.getConfiguration().getFormatter(formatterName);
		FormatterContext context = new FormatterContext(out, options.getFormatterProperties());
		return fd.getInstance(context);
	}

	private ColumnsHeader constructHeader(HeaderField... fields) throws FormatterException {
		try {
			RowSetMetaDataImpl metaData = new RowSetMetaDataImpl();
			metaData.setColumnCount(fields.length);

			for (int i = 0; i < fields.length; i++) {
				HeaderField hf = fields[i];
				int sqlIndex = i + 1;
				metaData.setColumnName(sqlIndex, hf.name);
				metaData.setColumnLabel(sqlIndex, hf.name);
				metaData.setColumnType(sqlIndex, hf.type.getCode());
				metaData.setColumnTypeName(sqlIndex, hf.type.name());
			}

			return new ColumnsHeader(metaData);
		} catch (SQLException e) {
			throw new FormatterException("Error while constructing table headers", e);
		}
	}

	private static class HeaderField {

		String name;
		SQLType type;

		public HeaderField(String name, SQLType type) {
			this.name = name;
			this.type = type;
		}
	}

	public enum InfoType {

		HELP {
			@Override
			public void showInfo(InfoLister infoLister) {
				infoLister.printResource(Constants.HELP_FILE);
			}
		},
		VERSION {
			@Override
			public void showInfo(InfoLister infoLister) {
				infoLister.printResource(Constants.VERSION_FILE);
			}
		},
		LICENSE {
			@Override
			public void showInfo(InfoLister infoLister) {
				infoLister.printResource(Constants.LICENSE_FILE);
			}
		},
		FORMATTERS {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.listFormatters();
			}
		},
		TYPES {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.listTypes();
			}
		},
		DATABASES {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.listDatabases();
			}
		},
		CONNECTION {
			@Override
			public void showInfo(InfoLister infoLister) throws FormatterException, ConfigurationException {
				infoLister.testConnection();
			}
		};

		public abstract void showInfo(InfoLister infoLister) throws ConfigurationException, FormatterException;
	}
}
